#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <byteswap.h>
#include "ttf.h"
#include "u8.h"
#include "etc.h"
#define DEBUG(x)

struct errorcontext {
	char *line;
	const char*context;
	unsigned *inp_ln;
};

void get_gsub_line(FILE *gsub_spec,struct errorcontext*e){
	if(!fgets(e->line,100,gsub_spec))
		printf_abort("%s ran out of line %u\n",e->context,*e->inp_ln);
	++*e->inp_ln;
}

void cant_parse_gsub_line(struct errorcontext*e){
	printf_abort("%s couldn't parse line %u:\n\t%s",e->context,*e->inp_ln,e->line);
}

void complain_gsub_line(struct errorcontext*e,const char *complaint){
	printf_abort("%s gsub line %u:\n%s\n\t%s",e->context,*e->inp_ln,complaint,e->line);
}

struct glyfmap {
	unsigned short from;
	unsigned short to;
};
#define MAXCONT 6
// data for an individual ligature
struct ligmap {
	unsigned short from;
	unsigned short n_comp;
	unsigned short cont[MAXCONT];
	unsigned short to;
	long lig_tbl_loc;
};
// data for a set of ligatures with the same first character
struct ligset {
	unsigned short from;
	unsigned int ligset_z;
	long ligset_tbl_loc;
	struct ligmap *ligs;
};

int glyfmap_compar(const unsigned short *a, const unsigned short *b){
	if(*a > *b)return 1;
	if(*a < *b)return -1;
	return 0;
}

struct feature {char tag[4]; int lookup;};

typedef int(*cmpfun)(const void*,const void*);

int feature_compar(struct feature*a, struct feature*b){
	return memcmp(a->tag,b->tag,4);
}

int getfrom(char *glyfs, size_t z, int i){
	return*(unsigned short*)(glyfs+z*i);
}

void write_coverage_table(FILE *o, char *g, size_t z, int n_glyphs){
	int glyphrange_count = 1;
	int prev_from = getfrom(g,z,0);
	int glyph_i;
	// count glyph ranges, to decide which table type to use
	for(glyph_i = 1; glyph_i < n_glyphs; glyph_i++) {
		if(getfrom(g,z,glyph_i) != prev_from + 1)glyphrange_count++;
		prev_from = getfrom(g,z,glyph_i);
	}
	if(glyphrange_count*3 > n_glyphs){
		write16(o, 1);//coverage table type 1
		write16(o, n_glyphs);
		for(glyph_i = 0; glyph_i < n_glyphs; glyph_i++) {
			write16(o, getfrom(g,z,glyph_i));
		}
	}else{
		write16(o, 2);//coverage table type 2
		write16(o, glyphrange_count);
		int prev_from = getfrom(g,z,0);
		int range_start = getfrom(g,z,0);
		int cov_start = 0;
		int cov_index = 1;
		for(glyph_i = 1; glyph_i < n_glyphs; glyph_i++) {
			if(getfrom(g,z,glyph_i) != prev_from + 1){
				write16(o,range_start);
				write16(o,prev_from);
				write16(o,cov_start);
				cov_start = cov_index;
				range_start = getfrom(g,z,glyph_i);
			}
			cov_index++;
			prev_from = getfrom(g,z,glyph_i);
		}
		write16(o,range_start);
		write16(o,prev_from);
		write16(o,cov_start);
	}
}

void write_ligtbl_placeholder(int ligset_z,FILE* out){
	write16(out,ligset_z);
	int lig_i;
	for(lig_i=0;lig_i<ligset_z;lig_i++){
		write16(out,0);
	}
}

void write_ligature_set(struct ligset* L, FILE* out){
	int ligset_z = L->ligset_z;
	int lig_i;
	for(lig_i=0;lig_i<ligset_z;lig_i++){
		L->ligs[lig_i].lig_tbl_loc = ftell(out);
		write16(out,L->ligs[lig_i].to);
		write16(out,L->ligs[lig_i].n_comp);
		int cont_i;
		for(cont_i=0;cont_i<L->ligs[lig_i].n_comp-1;cont_i++){
			write16(out,L->ligs[lig_i].cont[cont_i]);
		}
	}
}

void read_GSUB_subtable_4_1(FILE *gsub_spec, struct errorcontext *ectx, struct ligmap **mappings,\
	unsigned *n_glyphs, unsigned short *ccode2glyfid){
	int glyph_i = 0;
	char *line = ectx->line;
	get_gsub_line(gsub_spec,ectx);
	while(strcmp(line, "END MAPPINGS\n")){
		if(glyph_i >= *n_glyphs) {
			*n_glyphs*=2;
			*mappings=realloc(*mappings,*n_glyphs * sizeof(struct ligmap));
		}
		if(strlen(line)<3)cant_parse_gsub_line(ectx);
		const char *iter;
		int chr=u8sscanch(line,&iter);
		mappings[0][glyph_i].from=ccode2glyfid[chr];
		int ci = 0;
		chr = u8sscanch(iter, &iter);
		while(chr!='\t'&&chr!='\n'&&chr!=EOF){
			mappings[0][glyph_i].cont[ci]=ccode2glyfid[chr];
			chr = u8sscanch(iter,&iter);
			ci++;
		}
		mappings[0][glyph_i].n_comp=ci+1;
		if(chr!='\t')complain_gsub_line(ectx,"tab separator missing");
		chr=u8sscanch(iter,&iter);
		mappings[0][glyph_i].to=ccode2glyfid[chr];
		get_gsub_line(gsub_spec,ectx);
		glyph_i++;
      	}
	if(glyph_i < *n_glyphs) {
		*n_glyphs=glyph_i;
		*mappings=realloc(*mappings,*n_glyphs * sizeof(struct ligmap));
	}
}

void read_GSUB_subtable_1_2(FILE *gsub_spec, struct errorcontext *ectx,
	struct glyfmap **mappings,unsigned *n_glyphs, unsigned short *ccode2glyfid){
	int glyph_i=0;
	char *line = ectx->line;
	get_gsub_line(gsub_spec,ectx);
	while(strcmp(line, "END MAPPINGS\n")){
		if(glyph_i >= *n_glyphs) {
			*n_glyphs*=2;
			*mappings=realloc(*mappings,*n_glyphs * sizeof(struct glyfmap));
		}
		if(strlen(line)<3)cant_parse_gsub_line(ectx);
		const char *iter;
		int from_chr = u8sscanch(line, &iter);
		int sep = u8sscanch(iter, &iter);
		int to_chr = u8sscanch(iter, &iter);
		if(sep != '\t')
			complain_gsub_line(ectx,"tab separator missing");
		if(to_chr == EOF || from_chr == EOF )
			complain_gsub_line(ectx,"missing character");
		mappings[0][glyph_i].from = ccode2glyfid[from_chr];
		mappings[0][glyph_i].to = ccode2glyfid[to_chr];
		if(mappings[0][glyph_i].to == 0xffff || mappings[0][glyph_i].from == 0xffff)
			complain_gsub_line(ectx,"reference to unmapped character");
		get_gsub_line(gsub_spec,ectx);
		glyph_i++;
	}
	if(glyph_i < *n_glyphs) {
		*n_glyphs=glyph_i;
		*mappings=realloc(*mappings,*n_glyphs * sizeof(struct glyfmap));
	}

}

void write_GSUB_table(FILE *out,const char *gsub_file_name, unsigned short *ccode2glyfid) {
	struct gsub_spec *gs = read_GSUB_spec_txt(gsub_file_name,ccode2glyfid);
	write_gsub_spec_to_table(out,gs,ccode2glyfid);
}

struct gsub_lookup_subtable {
	unsigned type;
	unsigned subtype;
	unsigned n_glyphs;
	char *rawdata;
};

struct gsub_lookup_table {
	unsigned type, flags, n_subtables;
	struct gsub_lookup_subtable *subtables;
};

struct gsub_spec {
	unsigned n_scripts, n_features;
	struct feature *scripts;
	struct feature *features;
	struct gsub_lookup_table *lookups;
};

struct gsub_spec *read_GSUB_spec_txt(const char *gsub_file_name, unsigned short *ccode2glyfid) {
	FILE *gsub_spec = fopen(gsub_file_name, "r");
	if(!gsub_spec)printf_abort("couldn't open gsub file %s\n",gsub_file_name);
	char line[100]="";
	unsigned gsub_ln = 0;
	struct errorcontext ectx_here = {line,"while looking for scriptlist",&gsub_ln};
	struct errorcontext *ectx = &ectx_here;
	struct gsub_spec *gs = calloc(1,sizeof(struct gsub_spec));
	DEBUG(puts(ectx->context));
	while(strprfx(line,"SCRIPTLIST")){
		get_gsub_line(gsub_spec,ectx);
	}
	if(sscanf(line, "SCRIPTLIST %u", &gs->n_scripts)<1||gs->n_scripts==0){
		cant_parse_gsub_line(ectx);
	}
	gs->scripts = calloc(gs->n_scripts, sizeof(struct feature));
	char tag[5];
	int scrp_i;
	for(scrp_i=0;scrp_i<gs->n_scripts;scrp_i++){
		get_gsub_line(gsub_spec, ectx);
		if(sscanf(line, "SCRIPT %4s", tag)<1){
			cant_parse_gsub_line(ectx);
		}
		memcpy(gs->scripts[scrp_i].tag,tag,4);
	}
	ectx->context="while looking for lookuplist";
	DEBUG(puts(ectx->context));
	while(strprfx(line,"LOOKUPLIST")){
		get_gsub_line(gsub_spec,ectx);
	}
	if(sscanf(line, "LOOKUPLIST %u", &gs->n_features)<1){
		cant_parse_gsub_line(ectx);
	}
	DEBUG(printf("%u features found\n", gs->n_features));
	gs->features = calloc(gs->n_features, sizeof(struct feature));
	gs->lookups = calloc(gs->n_features, sizeof(struct gsub_lookup_table));
        int lookup_index;
	for(lookup_index=0;lookup_index<gs->n_features;lookup_index++){
		ectx->context = "while looking for lookup header";
		DEBUG(puts(ectx->context));
		get_gsub_line(gsub_spec, ectx);
		unsigned lookuptype, lookupflags, subtablecount;
		if(sscanf(line, "LOOKUP %d %x %u", &lookuptype,
			&lookupflags, &subtablecount)<3){
			cant_parse_gsub_line(ectx);
		}
		gs->lookups[lookup_index].type = lookuptype;
		gs->lookups[lookup_index].flags = lookupflags;
		gs->lookups[lookup_index].n_subtables = subtablecount;
		ectx->context = "while looking for feature tag";
		DEBUG(puts(ectx->context));
		get_gsub_line(gsub_spec, ectx);
		if(sscanf(line, "FEATURE %4s", tag)<1){
			cant_parse_gsub_line(ectx);
		}
		memcpy(gs->features[lookup_index].tag,tag,4);
		gs->features[lookup_index].lookup = lookup_index;
		gs->lookups[lookup_index].subtables \
			= calloc(subtablecount, sizeof(struct gsub_lookup_subtable));
		int subtableindex;
		for(subtableindex=0;subtableindex<subtablecount;subtableindex++){
			struct gsub_lookup_subtable *glsb \
				= &gs->lookups[lookup_index].subtables[subtableindex];
			ectx->context = "while looking for subtable type";
			DEBUG(puts(ectx->context));
			get_gsub_line(gsub_spec, ectx);
			unsigned subtabletype;
			if(sscanf(line, "SUBTABLE %d", &subtabletype)<1){
				cant_parse_gsub_line(ectx);
			}
			glsb->type = lookuptype;
			glsb->subtype = subtabletype;
			ectx->context = "while looking for mappings header";
			get_gsub_line(gsub_spec, ectx);
			if(sscanf(line,"MAPPINGS %d",&glsb->n_glyphs)<1){
				cant_parse_gsub_line(ectx);
			}
			if(lookuptype == 4 && subtabletype == 1){
				glsb->rawdata = calloc(glsb->n_glyphs,sizeof(struct ligmap));
				ectx->context = "while reading mappings for subtable 4.1";
				read_GSUB_subtable_4_1(gsub_spec,ectx,&glsb->rawdata,&glsb->n_glyphs,ccode2glyfid);
			}else if(lookuptype == 1 && subtabletype == 2){
				glsb->rawdata = calloc(glsb->n_glyphs,sizeof(struct glyfmap));
				ectx->context = "while reading mappings for subtable 1.2";
				read_GSUB_subtable_1_2(gsub_spec,ectx,&glsb->rawdata,&glsb->n_glyphs,ccode2glyfid);
			}else{
				complain_gsub_line(ectx,"unknown lookup subtype");
			}
			if(strcmp(line, "END MAPPINGS\n")){
				complain_gsub_line(ectx,"END MAPPINGS expected");
			}
		}
	}
	get_gsub_line(gsub_spec, ectx);
	if(strcmp(line, "END LOOKUPS\n")){
		complain_gsub_line(ectx,"END LOOKUPS expected");
	}
	qsort(gs->features,gs->n_features,sizeof(struct feature),feature_compar);
	fclose(gsub_spec);
	return gs;
}

void write_gsub_subtable_spec_1_2(FILE *out, unsigned n_glyphs, struct glyfmap *mappings, unsigned short *ccode2glyfid);
void write_gsub_subtable_spec_4_1(FILE *out, unsigned n_glyphs, struct ligmap *mappings, unsigned short *ccode2glyfid);
void write_gsub_lookupspec(FILE *out,struct gsub_spec *gs, unsigned short *ccode2glyfid);

void write_gsub_spec_to_table(FILE *out,struct gsub_spec *gs, unsigned short *ccode2glyfid) {
	unsigned gsub_start = ftell(out);
	write16(out,1);
	write16(out,1);
	write16(out,0);//script list offset
	write16(out,0);//feature list offset
	write16(out,0);//lookup list offset
	write32(out,0);//feature variations offset
	unsigned lookup_table_start = ftell(out);
	unsigned scrp_i;
	write_gsub_lookupspec(out,gs,ccode2glyfid);
	unsigned scriptlist_table_start = ftell(out);
	long script_table_offset = 2 + gs->n_scripts * 6;
	write16(out,gs->n_scripts);//script count
	for(scrp_i=0;scrp_i<gs->n_scripts;scrp_i++){
		writemem(out,gs->scripts[scrp_i].tag,4);
		write16(out,script_table_offset);
	}
	write16(out,4);//offset to default langsys table
	write16(out,0);//no other languages besides the default
	//start of langsys
	write16(out,0);//reserved
	write16(out,0xffff);//required feature index
	write16(out,gs->n_features);
	unsigned fi;
	for(fi=0;fi<gs->n_features;fi++){
		write16(out,fi);
	}
	unsigned featurelist_table_start = ftell(out);
	write16(out,gs->n_features);//feature count
	for(fi=0;fi<gs->n_features;fi++){
		writemem(out,gs->features[fi].tag,4);
		write16(out,2+gs->n_features*6+fi*6);
	}
	for(fi=0;fi<gs->n_features;fi++){
		write16(out,0);
		write16(out,1);
		write16(out,gs->features[fi].lookup);
	}
	goloc(out,gsub_start+4);
	write16(out,scriptlist_table_start-gsub_start);
	write16(out,featurelist_table_start-gsub_start);
	write16(out,lookup_table_start-gsub_start);
	ffastfw(out);
}

void write_gsub_lookupspec(FILE *out,struct gsub_spec *gs, unsigned short *ccode2glyfid) {
	unsigned lookup_table_start = ftell(out);
	write16(out, gs->n_features);
	unsigned lookup_index;
	long lookup_list_offset = ftell(out);
	for(lookup_index=0;lookup_index<gs->n_features;lookup_index++) write16(out,0);
	for(lookup_index=0;lookup_index<gs->n_features;lookup_index++){
		struct gsub_lookup_table *gslt = gs->lookups + lookup_index;
		long lookup_loc = ftell(out);
		goloc(out,lookup_list_offset+2*lookup_index);
		write16(out,lookup_loc-lookup_table_start);
		ffastfw(out);
		write16(out, gslt->type);
		write16(out, gslt->flags);
		write16(out, gslt->n_subtables);
		unsigned subtableindex;
		long subtable_list_loc = ftell(out);
		// write out subtable offset placeholders
		for(subtableindex=0;subtableindex<gslt->n_subtables;subtableindex++)
			write16(out,0);
		for(subtableindex=0;subtableindex<gslt->n_subtables;subtableindex++){
			struct gsub_lookup_subtable *gsbt = gslt->subtables + subtableindex;
			long subtable_loc = ftell(out);
			if(gsbt->type == 1 && gsbt->subtype == 2){
				write_gsub_subtable_spec_1_2(out,gsbt->n_glyphs, (struct glyfmap*)gsbt->rawdata,ccode2glyfid);
			}else if(gsbt->type == 4 && gsbt->subtype == 1){
				write_gsub_subtable_spec_4_1(out,gsbt->n_glyphs, (struct ligmap*)gsbt->rawdata,ccode2glyfid);
			}else{
				printf_abort("unknown subtable type ltb %d sb %d typ %d subtyp %d\n",
					lookup_index,subtableindex,gsbt->type,gsbt->subtype);
			}
			goloc(out,subtable_list_loc+2*subtableindex);
			write16(out,subtable_loc-lookup_loc);
			ffastfw(out);
		}
	}
}

void write_gsub_subtable_spec_1_2(
	FILE *out,unsigned n_glyphs,
	struct glyfmap*mappings,
	unsigned short*ccode2glyfid
){
	unsigned subtable_loc= ftell(out);
	write16(out, 2);
	long coveragetable_offset_loc= ftell(out);
	// write out the coverage table offset placeholder
	write16(out,0);
	unsigned glyph_i;
	// sort mappings
	bublesort((char*)mappings,n_glyphs,sizeof(struct glyfmap),(cmpfun)glyfmap_compar);
	write16(out, n_glyphs);
	// write out the glyphs to map to
	unsigned short lastfrom = mappings[0].from;
	for(glyph_i=0;glyph_i<n_glyphs;glyph_i++){
		write16(out, mappings[glyph_i].to);
		lastfrom = mappings[glyph_i].from;
	}
	long coveragetable_loc= ftell(out);
	write_coverage_table(out,(char*)mappings,sizeof(struct glyfmap),n_glyphs);
	goloc(out,coveragetable_offset_loc);
	write16(out,coveragetable_loc-subtable_loc);
	ffastfw(out);
}

void write_gsub_subtable_spec_4_1(FILE *out, unsigned n_glyphs, struct ligmap *mappings, unsigned short *ccode2glyfid){
	unsigned subtable_loc = ftell(out);
	write16(out, 1);
	long coveragetable_offset_loc = ftell(out);
	write16(out, 0);//coverage table offset
	unsigned glyph_i;
	// sort ligatures by first character (TODO: use a stable sort instead)
	qsort(mappings,n_glyphs,sizeof(struct ligmap),(cmpfun)glyfmap_compar);
	// count how many ligature sets there are
	unsigned n_ligsets = 1;
	unsigned short curfrom = mappings[0].from;
	for(glyph_i = 1; glyph_i < n_glyphs; glyph_i++) {
		if(mappings[glyph_i].from != curfrom){
			curfrom = mappings[glyph_i].from;
			n_ligsets++;
		}
	}
	struct ligset *ligsets = malloc(sizeof(struct ligset) * n_ligsets);
	write16(out, n_ligsets); // n_ligsets
	long ligset_table_offsets_loc = ftell(out);
	unsigned ligset_i;
	unsigned lig_i;
	for(ligset_i=0;ligset_i<n_ligsets;ligset_i++){
		write16(out, 0);
	}
	ligset_i=0;
	curfrom=mappings[0].from;
	long ligset_tables_loc = ftell(out);
	unsigned ligset_z=0;
	// divide the ligatures into sets based on first character
	ligsets[0].ligs = mappings;
	for(glyph_i=0;glyph_i<=n_glyphs;glyph_i++){
		if(glyph_i==n_glyphs||mappings[glyph_i].from!=curfrom){
			ligsets[ligset_i].from = curfrom;
			ligsets[ligset_i].ligset_z = ligset_z;
			ligsets[ligset_i].ligset_tbl_loc = ftell(out);
			ligsets[ligset_i].ligs = mappings + glyph_i - ligset_z;
			write_ligtbl_placeholder(ligset_z,out);
			ligset_z=0;
			ligset_i++;
			if(glyph_i!=n_glyphs)
				curfrom = mappings[glyph_i].from;
		}
		ligset_z++;
	}
	for(ligset_i=0;ligset_i<n_ligsets;ligset_i++){
		write_ligature_set(ligsets+ligset_i,out);
	}
	// go to each placeholder lig table and fill in offsets
	for(ligset_i=0;ligset_i<n_ligsets;ligset_i++){
		ligset_z = ligsets[ligset_i].ligset_z;
		long ligset_tbl_loc = ligsets[ligset_i].ligset_tbl_loc;
		goloc(out,ligset_tbl_loc + 2);
		for(lig_i=0;lig_i<ligset_z;lig_i++){
			write16(out,ligsets[ligset_i].ligs[lig_i].lig_tbl_loc-ligset_tbl_loc);
		}
	}
	// write out offsets to ligature tables
	goloc(out,ligset_table_offsets_loc);
	for(ligset_i=0;ligset_i<n_ligsets;ligset_i++){
		write16(out,ligsets[ligset_i].ligset_tbl_loc - subtable_loc);
	}
	ffastfw(out);
	long covtable_loc = ftell(out);
	write_coverage_table(out,(char*)ligsets, sizeof(struct ligset), n_ligsets);
	goloc(out,coveragetable_offset_loc);
	write16(out,covtable_loc - subtable_loc);
	ffastfw(out);
	free(ligsets);
}

