#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include "bdf.h"
#include "u8.h"
#include "unidat.h"
#include "etc.h"

/*

STARTFONT 2.1
FONT neoletters
SIZE 12 75 75
FONTBOUNDINGBOX 18 16 0 -4
STARTPROPERTIES 3
PIXEL_SIZE 16
FONT_ASCENT 12
FONT_DESCENT 4
ENDPROPERTIES
CHARS 21576
*/

#define MAX_FONT_PROPS 40
#define MAX_GLYPH 0x110000
#define FULLWIDTH 18
#define BITFIELDH 16

typedef uint32_t bitline;

struct bdfglyph {
	char name[16];
	int encoding;
	int dwidth;
	int bbx[4];
	bitline bitmap[BITFIELDH];
};

struct bdfinfo {
	char name[64];
	int dpi;
	int size;
	unsigned halfwidth;
	int bbx[4];
	unsigned nprops;
	char **prop_lines;
	unsigned nchars;
};

struct bdffont {
	struct bdfinfo i;
	struct bdfglyph**g;
	unsigned glyph_slots;
};

struct bdfglyph *make_blank_char(unsigned gi) {
	struct bdfglyph *g = calloc(1,sizeof(struct bdfglyph));
	if(!g)return g;
	g->encoding = gi;
	sprintf(g->name,"uni%X",gi);
	g->dwidth = get_eaw(gi)*9;
	return g;
}

struct bdfglyph *copy_char(unsigned gi, struct bdfglyph *gc) {
	struct bdfglyph *g = malloc(sizeof(struct bdfglyph));
	if(!g)return g;
	*g = *gc;
	g->encoding = gi;
	sprintf(g->name,"uni%X",gi);
	g->dwidth = get_eaw(gi)*9;
	return g;
}

void calc_bbx(struct bdfglyph *g){
	int j;
	unsigned orsum=0;
	int topnonzero = -1;
	int botnonzero = -1;
	for(j=0;j<BITFIELDH;j++){
		orsum |= g->bitmap[j];
		if(topnonzero == -1 && g->bitmap[j] != 0) topnonzero = j;
		if(g->bitmap[j] != 0) botnonzero = j;
	}
	int height = botnonzero - topnonzero + 1;
	int yoff = 11 - botnonzero;
	if (botnonzero == -1)yoff = 0;
	int i;
	int lefnonzero = -1;
	int rignonzero = -1;
	for(i=0;i<FULLWIDTH;i++){
		int p = !!(((1<<(FULLWIDTH-1))>>i)&orsum);
		if(p) rignonzero = i;
		if(p && lefnonzero == -1) lefnonzero = i;
	}
	int width = rignonzero - lefnonzero + 1;
	int xoff = lefnonzero;
	if(lefnonzero == -1)xoff = 0;
	g->bbx[0] = width;
	g->bbx[1] = height;
	g->bbx[2] = xoff;
	g->bbx[3] = yoff;
}

struct io_info {
	int shift; // left for input, right for output
	int top; // first row to output
	int bot; // one after last row to output
	int bytes; // bytes to output
};

struct io_info calc_io_info(struct bdfglyph *g){
	int width = g->bbx[0];
	int height = g->bbx[1];
	int xoff = g->bbx[2];
	int yoff = g->bbx[3];
	int bwidth = width/8 + !!(width%8);
	struct io_info o;
	o.bot = 12 - yoff;
	o.top = o.bot - height;
	o.shift = FULLWIDTH - bwidth * 8 - xoff;
	o.bytes = bwidth;
	return o;
}

struct bdffont *new_bdffont(){
	struct bdffont*font=malloc(sizeof(struct bdffont));
	if(!font)return font;
	strncpy(font->i.name,"",64);
	font->i.dpi=75;
	font->i.size=12;
	memcpy(font->i.bbx,(int[]){FULLWIDTH,BITFIELDH,0,-4},sizeof(int)*4);
	font->i.nprops=0;
	font->i.prop_lines=0;
	font->g=(void*)calloc(MAX_GLYPH,sizeof(void*));
	font->glyph_slots=MAX_GLYPH;
	return font;
}

void copy_font_info(struct bdffont *to, struct bdffont *from){
	strncpy(to->i.name,from->i.name, 64);
	to->i.dpi=from->i.dpi;
	memcpy(to->i.bbx,from->i.bbx,sizeof(int)*4);
	to->i.size=from->i.size;
	if(to->i.prop_lines)free(to->i.prop_lines);
	to->i.prop_lines=malloc(from->i.nprops*sizeof(char*));
	to->i.nprops=from->i.nprops;
	for(int i=0;i<to->i.nprops;i++){
		const char*s=from->i.prop_lines[i];
		to->i.prop_lines[i]=strcpy(malloc(strlen(s)+1),s);
	}
}

void load_font(const char * filename, struct bdfglyph **font, struct bdfinfo *info, void(*status_callback)(struct bdfglyph *g, int cur, int tot)){
	FILE *bdfin = fopen(filename,"r");
	if(!bdfin){
		fprintf(stderr,"unable to open file %s\r\n",filename);
		abort();
	}
	char line[300];
	do {
		if(!fgets(line,300,bdfin))return;
		chomp(line);
		if (!strprfx(line,"FONT "))strcpy(info->name,line+5);
		if (!strprfx(line,"SIZE "))sscanf(line,"SIZE %d %d",&info->size,&info->dpi);
		if (!strprfx(line,"FONTBOUNDINGBOX "))sscanf(line,"FONTBOUNDINGBOX %d %d %d %d",info->bbx,info->bbx+1,info->bbx+2,info->bbx+3);
		if (!strprfx(line,"STARTPROPERTIES ")){
			sscanf(line,"STARTPROPERTIES %d",&info->nprops);
			info->prop_lines = malloc(sizeof(char*)*MAX_FONT_PROPS);
			unsigned pi=0;// properties iterator.
			do {
				if(!fgets(line,300,bdfin))return;
				chomp(line);
				if(!strprfx(line,"ENDPROPERTIES"))break;
				info->prop_lines[pi] = malloc(strlen(line)+1);
				strcpy(info->prop_lines[pi],line);
				pi++;
			}while(pi < info->nprops);
		}
		if (!strprfx(line,"CHARS ")) break;
	} while(1);

	sscanf(line,"CHARS %d",&info->nchars);
	int count = 0;
	do {
		struct bdfglyph g;
		if(!fgets(line,300,bdfin))break;
		if(!strprfx(line,"ENDFONT"))break;
		sscanf(line,"STARTCHAR %16s",g.name);
		if(!fgets(line,300,bdfin))break;
		int ccode;
		sscanf(line,"ENCODING %d",&g.encoding);
		ccode = g.encoding;

		do if(!fgets(line,300,bdfin))break;
		while(strprfx(line,"DWIDTH"));
		sscanf(line,"DWIDTH %d",&g.dwidth);
		if(!fgets(line,300,bdfin))break;
		sscanf(line,"BBX %d %d %d %d",g.bbx,g.bbx+1,g.bbx+2,g.bbx+3);
		if(!fgets(line,300,bdfin))break;
		sscanf(line,"BITMAP");
		struct io_info O = calc_io_info(&g);
		int j;
		for(j=0;j<BITFIELDH;j++){
			g.bitmap[j]=0;
			if(j >= O.top && j < O.bot) {
				if(!fgets(line,300,bdfin))break;
				unsigned x;
				sscanf(line,"%x",&x);
				g.bitmap[j] = leftshft(x,O.shift);
			}
		}
		if(status_callback)status_callback(&g,count,info->nchars);
		font[ccode] = (struct bdfglyph*)malloc(sizeof(struct bdfglyph));
		*font[ccode] = g;
		do j = !!fgets(line,300,bdfin);
		while(j && strprfx(line,"ENDCHAR"));
		count++;
	}while(1);
	fclose(bdfin);

}

void save_font(const char * filename, struct bdfglyph **font, struct bdfinfo *info, void(*status_callback)(struct bdfglyph *g, int cur, int tot)){
	FILE *bdfout = fopen(filename,"w");
	unsigned gi;
	unsigned charcount = 0; //count the characters in font
	for (gi = 0; gi < MAX_GLYPH; gi++) {
		if(font[gi])charcount++;
	}
	fprintf(bdfout,"STARTFONT 2.1\n");
	fprintf(bdfout,"FONT %s\n",info->name);
	fprintf(bdfout,"SIZE %d %d %d\n", info->size, info->dpi, info->dpi);
	fprintf(bdfout,"FONTBOUNDINGBOX %d %d %d %d\n", info->bbx[0], info->bbx[1], info->bbx[2], info->bbx[3]);
	fprintf(bdfout,"STARTPROPERTIES %d\n", info->nprops);
	unsigned pi = 0;
	while(pi < info->nprops) {
		fprintf(bdfout,"%s\n",info->prop_lines[pi]);
		pi++;
	}
	fprintf(bdfout,"ENDPROPERTIES\n");
	fprintf(bdfout,"CHARS %d\n",charcount);
	int outcount = 0;
	for (gi = 0; gi < MAX_GLYPH; gi++) {
		struct bdfglyph *g = font[gi];
		if(!g)continue;
		g->dwidth = get_eaw(gi)*9;
		if(status_callback)status_callback(g,outcount,charcount);
		fprintf(bdfout,"STARTCHAR %s\n",g->name);
		fprintf(bdfout,"ENCODING %d\n",gi);
		fprintf(bdfout,"SWIDTH %d 0\n",g->dwidth*500/8);
		fprintf(bdfout,"DWIDTH %d 0\n",g->dwidth);
		calc_bbx(g);
		fprintf(bdfout,"BBX %d %d %d %d\nBITMAP\n",g->bbx[0],g->bbx[1],g->bbx[2],g->bbx[3]);
		struct io_info O = calc_io_info(g);
		int j;
		for(j=O.top;j<O.bot;j++){
			fprintf(bdfout,"%0*X\n",O.bytes*2,leftshft(g->bitmap[j],-O.shift));
		}
		fprintf(bdfout,"ENDCHAR\n");
		outcount++;
	}
	fprintf(bdfout,"ENDFONT\n");
	fclose(bdfout);
}

/* 4 lines tall */
void outputasbitmap(struct bdfglyph *g,int y, int x){
	int j;
	for(j=0;j<BITFIELDH/4;j++){
		int b;
		printf("\033[%d;%dH",y+j,x);
		for (b=(1<<(FULLWIDTH-1));b!=0;b/=4) {
			int t =   (g->bitmap[j*4]&b	?0x01:0x00)
				| (g->bitmap[j*4]&(b/2)	?0x08:0x00)
				| (g->bitmap[j*4+1]&b  	?0x02:0x00)
				| (g->bitmap[j*4+1]&(b/2)?0x10:0x00)
				| (g->bitmap[j*4+2]&b	?0x04:0x00)
				| (g->bitmap[j*4+2]&(b/2)?0x20:0x00)
				| (g->bitmap[j*4+3]&b    ?0x40:0x00)
				| (g->bitmap[j*4+3]&(b/2)?0x80:0x00);
			printf("\342%c%c",(t>>6)|0240,(t&077)|0200);
		}
	}
}

struct bdffont *load_font_alloc(const char * filename, void(*status_callback)(struct bdfglyph *g, int cur, int tot)) {
	struct bdfglyph **g = (void*)calloc(MAX_GLYPH,sizeof(void*));
	struct bdffont *font = malloc(sizeof(struct bdffont));
	if(!g||!font){
		if(g)free(g);
		if(font)free(font);
		return NULL;
	}
	load_font(filename, g, &font->i, status_callback);
	font->g = g;
	font->glyph_slots = MAX_GLYPH;
	return font;
}

void save_font_struct(const char * filename, struct bdffont *font, void(*status_callback)(struct bdfglyph *g, int cur, int tot)){
	save_font(filename, font->g, &font->i, status_callback);
}

const char *get_font_name(struct bdffont *font) {
	return font->i.name;
}

const int *get_font_bbx(struct bdffont *font) {
	return font->i.bbx;
}

unsigned font_nchars(struct bdffont *font){
	return font->i.nchars;
}

void free_bdffont(struct bdffont *font){
	unsigned i;
	for(i=0;i<font->glyph_slots;i++){
		if(font->g[i])free(font->g[i]);
	}
	free(font->g);
	for(i=0;i < font->i.nprops;i++){
		free(font->i.prop_lines[i]);
	}
	free(font->i.prop_lines);
	free(font);
}

struct bdfglyph *get_glyph(struct bdffont *font, unsigned i) {
	if(i > font->glyph_slots)abort();
	return font->g[i];
}

struct bdfglyph **get_glyph_array(struct bdffont *font) {
	return font->g;
}

unsigned font_glyph_slots(struct bdffont *font){
	return font->glyph_slots;
}

void set_glyph(struct bdffont *font, unsigned i, struct bdfglyph *g){
	if(i > font->glyph_slots)abort();
	if(font->g[i] != NULL){
		free(font->g[i]);
	}
	font->g[i] = g;
	g->encoding = i;
}

void edit_glyph_pixel(struct bdfglyph *g, int y, int x, int op) {
	g->bitmap[y] &= ~(0x20000 >> x);
	if(op)g->bitmap[y] |= (0x20000 >> x);
}

int get_glyph_pixel(struct bdfglyph *g, int y, int x) {
	if(x<0||y<0||x>=FULLWIDTH||y>=BITFIELDH)return 0;
	return !!(g->bitmap[y] & (0x20000 >> x));
}

void shift_glyph_bitmap(struct bdfglyph *g, char direction) {
	if (g == NULL) return;
	unsigned i;
	if (direction == 'L') {
		for(i = 0; i < BITFIELDH; i++)g->bitmap[i] = (g->bitmap[i] << 1) & 0x3FFFF;
	} else if (direction == 'R') {
		for(i = 0; i < BITFIELDH; i++)g->bitmap[i] = g->bitmap[i] >> 1;
	} else if (direction == 'U') {
		for(i = 0; i < BITFIELDH-1; i++)g->bitmap[i] = g->bitmap[i+1];
		g->bitmap[BITFIELDH-1] = 0;
	} else if (direction == 'D') {
		for(i = BITFIELDH-1; i >= 1; i--)g->bitmap[i] = g->bitmap[i-1];
		g->bitmap[0] = 0;
	} else {
		abort();
	}
}

unsigned get_glyph_encoding(struct bdfglyph *g) {
	return g->encoding;
}

const char *get_glyph_name(struct bdfglyph *g) {
	return g->name;
}

int glyph_bitwise_cmp(struct bdfglyph *a, struct bdfglyph *b) {
	if(a == NULL && b == NULL)return 0;
	if(a == NULL && b != NULL)return -1; //null before nonnull
	if(a != NULL && b == NULL)return 1; //null before nonnull
	int j;
	for (j=0;j<BITFIELDH;j++){
		if(a->bitmap[j] < b->bitmap[j])return -1;
		if(a->bitmap[j] > b->bitmap[j])return 1;
	}
	if(a->dwidth < b->dwidth)return -1;
	if(a->dwidth > b->dwidth)return 1;
	return 0;
}
