#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "u8.h"
#include "bdf.h"
#include "oio.h"
#include "unidat.h"

/* editor for editing my neoletters font.
Written in 2019-2022 by Oren Watson
*/

#define ei else if
#define WIDTH 9


const char * bdfeditlogo = "𝔅𝔇𝔉𝕖𝕕𝕚𝕥";


void show_uni_name(unsigned ch) {
	const char *s = get_unidat_entry(ch);
	if (!s) {
		printf("<N/A>\r\n");
	} else {
		int len = (char*)memchr(s,';',100)-s;
		printf("%.*s\r\n",len,s);
	}
}

void show_help();

void showbar(int total, int cur, int wid) {
	if (total == 0)return;
	// full block 022610 -> one eighth block 022617 -> space 040
	int pix = cur * 8*wid / total;
	int fb = pix / 8;
	int px = pix % 8;
	int i = 0;
	while(i++<fb)fputs("\342\226\210",stdout);
	if(px!=0)printf("\342\226%c",0220-px);
	int bl = wid - fb - !!px;
	i = 0;
	while(i++<bl)putchar(' ');
}


void outputbigwithcur(struct bdfglyph *g, int y, int x){ //
	const char *const spa = "  ";
	const char *const blk = "██";
	const char *const spacur = "\033[32m▒▒\033[0m";
	const char *const blkcur = "\033[42m▒▒\033[0m";
	int j;
	int pixwid = get_eaw(get_glyph_encoding(g)) * WIDTH;
	if(pixwid == 0) pixwid = WIDTH;
	for(j = 0;j < 16;j++) {
		int i;
		fputs("\e[0m║",stdout);
		for (i=0;i<18;i++) {
			if (i == pixwid)fputs("\e[44m",stdout);
			if (i == x && j == y) {
				if(get_glyph_pixel(g,j,i)) fputs(blkcur,stdout);
				else fputs(spacur,stdout);
			} else {
				if(get_glyph_pixel(g,j,i))fputs(blk,stdout);
				else fputs(spa,stdout);
			}
		}
		fputs("\e[0m║\n\r",stdout);
	}
}

void bdfedit_status(struct bdfglyph *g,int cur,int tot) {
	printf("\e[7H\e[Kstatus: U%05x\n\r\e[Kname: %s",get_glyph_encoding(g),get_glyph_name(g));
	printf("\e[2H%3d%%\e[36;43m",cur*100/tot);
	showbar(tot,cur,30);
	printf("\e[0m");
	outputasbitmap(g,3,1);
}

#define HISTSZ 10
struct char_hist{
	unsigned hist[HISTSZ];
	int i;
};

unsigned *get_hist(struct char_hist *hist, int s){
	return &(hist->hist[(hist->i+s)%HISTSZ]);
}

void add_to_hist(struct char_hist *hist, unsigned ch){
	int i;
	for(i=0; i<HISTSZ; i++)if(*get_hist(hist,i)==ch)return;
	// if the character is not in history already, add it.
	hist->i--;
	if(hist->i<0)hist->i += HISTSZ;
	hist->hist[hist->i] = ch;
}

void show_hist(struct bdffont *font, unsigned ci, struct char_hist *hist, 
		unsigned gi, unsigned x, unsigned selected){
	printf(CSI "2;%dH clipboard\n\r" CLRCOL,x);
	printf(CSI "3;%dH %05x ",x,ci);
	show_uni_name(ci);
	printf(CSI "4;%dH history\n\r" CLRCOL,x);
	printf(CSI "5;%dH(%c) %05x ",x,selected==gi?'x':' ',gi);
	show_uni_name(gi);
	int c;
	for (c = 0; c < HISTSZ; c++) {
		unsigned hi = *get_hist(hist, c);
		printf(CSI "%d;%dH(%c) %05x ",6+c,x,selected==hi?'x':' ',hi);
		show_uni_name(hi);
	}
}

int are_you_sure(){
	printf(CLRSCR TOPLEF BBLU "%s - EXIT WITHOUT SAVING? (Y/N)\n\r" CLRCOL,bdfeditlogo);
	while(1){
		int c = oio_getkey();
		if(c == 'y' || c == 'Y')return 1;
		if(c == 'n' || c == 'N')return 0;
	}
}

struct editor_state {
	unsigned gi;
	unsigned ci;
	int y,x;
	int redraw;
	struct char_hist hist;
	const char *filename;
};

void quik_help();

void redraw_screen(struct editor_state *ed, struct bdffont *font){
	printf(CLRSCR TOPLEF "\033[44;96m%s %s\n\r" CLRCOL,bdfeditlogo,ed->filename);
	struct bdfglyph *g = get_glyph(font,ed->gi);
	if (g) {
		outputasbitmap(g,2,1);
		printf("\033[3;15H◌%s",u8chtostr(ed->gi));
		printf("\033[6;1H");
		printf("character %x\r\n",ed->gi);
		printf("unicode name ");
		show_uni_name(ed->gi);
		outputbigwithcur(g,ed->y,ed->x);
	} else {
		printf ("no character U%x\n\r",ed->gi);
		show_uni_name(ed->gi);
		printf ("(^A to add blank)");
	}
	quik_help();
	show_hist(font, ed->ci, &ed->hist, ed->gi, 40, ed->gi);
}

void show_select_hist(struct bdffont *font, struct editor_state *ed){
	printf(CLRSCR TOPLEF BBLU "%s - char select history\n\r" CLRCOL,bdfeditlogo);
	int han = -1;
	while(1) {
		show_hist(font, ed->ci, &ed->hist,ed->gi,0,han==-1?ed->gi:*get_hist(&ed->hist,han));
		int c = oio_getkey();
		if(c == KEY_UP && han != -1)han--;
		ei(c == KEY_DOWN && han-1 < HISTSZ-1)han++;
		ei(c == KEY_ENTER && han >= 0){
			//swap the character with the one from the history.
			unsigned *x = get_hist(&ed->hist, han);
			int y = ed->gi;
			ed->gi = *x;
			*x = y;
			return;
		}ei(c == KEY_BKSP || c == KEY_DEL){
			return;
		}
	}
}

int main(int argc, char **argv){
	if(argc < 2) {
		printf("need a BDF to edit\n");
		return 0;
	}
	char filename[100];
	strcpy(filename,argv[1]);

	oio_setup();

 	printf(CLRSCR TOPLEF BBLU "%s - loading unicode data\n\r" CLRCOL,bdfeditlogo);
	if(load_unidat()){
		oio_finish();
		printf("load_unidat failed\n");
		exit(1);
	}

	if(load_eawdat()){
		oio_finish();
		printf("load_eawdat failed\n");
		exit(1);
	}

	printf(CLRSCR TOPLEF BBLU "%s - loading %s\n\r" CLRCOL,bdfeditlogo,filename);

	struct bdffont *font = load_font_alloc(filename,bdfedit_status);

	/* editor state */
	struct editor_state ed = {60,60,0,0,1,{{'A','b','c','d','e','f','g','h','i','j'},0},filename};
	unsigned max_glyph = font_glyph_slots(font);
	while (1) {
		if(ed.gi >= max_glyph)ed.gi = max_glyph-1;
		if(ed.redraw)redraw_screen(&ed,font);
		int c = oio_getkey();
		ed.redraw = 1;
		if (c == CTRL('Q')) {
			printf("\033[2J\033[H\033[44m%s - save+quit %s\n\r\033[0m",bdfeditlogo,filename);
			save_font_struct(filename,font,bdfedit_status);
			break; // quit
		}ei (c == CTRL('S')) {
			printf(CLRSCR TOPLEF BBLU "%s - saving %s\n\r" CLRCOL,bdfeditlogo,filename);
			save_font_struct(filename,font,bdfedit_status);
		} ei (c == CTRL('X')){
			if(are_you_sure())break;
		} ei (c == CTRL('C'))ed.ci = ed.gi;
		ei (c == CTRL('G')){
			int c;
			printf(GOLN(2)CLRLIN);
			unsigned oldgi = ed.gi;
			do {
				c = oio_gethex(2,0,"go to char:",5,&ed.gi);
			} while(c != KEY_ENTER && c != '\n');
			if(ed.gi != oldgi){
				add_to_hist(&ed.hist,oldgi);
			}
		} ei (c == KEY_F1)show_help();
		ei (c == CTRL('H') || c == 'H'){
			show_select_hist(font, &ed);
		} ei (c == CTRL('A')){
			if (get_glyph(font,ed.gi) == NULL) {
				set_glyph(font,ed.gi,make_blank_char(ed.gi));
			}
		}ei (c == CTRL('K'))set_glyph(font,ed.gi,NULL);
		ei (c == CTRL('V')) {
			set_glyph(font,ed.gi,copy_char(ed.gi,get_glyph(font,ed.ci)));
		}
		ei (c == CTRL('W')) {
			struct bdfglyph *c = get_glyph(font,ed.ci);
			struct bdfglyph *g = get_glyph(font,ed.gi);
			struct bdfglyph *x = copy_char(ed.gi,c);
			free(c);
			set_glyph(font,ed.ci,copy_char(ed.ci,g));
			free(g);
			set_glyph(font,ed.gi,x);
		}
		ei (c == 'j') {
			shift_glyph_bitmap(get_glyph(font,ed.gi),'L');
		}
		ei (c == 'l') {
			shift_glyph_bitmap(get_glyph(font,ed.gi),'R');
		}
		ei (c == 'i') {
			shift_glyph_bitmap(get_glyph(font,ed.gi),'U');
		}
		ei (c == 'k') {
			shift_glyph_bitmap(get_glyph(font,ed.gi),'D');
		}
		ei (c == KEY_UP && ed.y > 0)ed.y--;
		ei (c == KEY_DOWN && ed.y < 15)ed.y++;
		ei (c == KEY_LEFT && ed.x > 0)ed.x--;
		ei (c == KEY_RIGHT && ed.x < 17)ed.x++;
		ei (c == KEY_PGUP && ed.gi > 0)ed.gi--;
		ei (c == KEY_PGDN && ed.gi < max_glyph-1)ed.gi++;
		ei (c == KEY_HOME) {
			if (ed.gi > 0)ed.gi--;
			while(ed.gi > 0 && !get_glyph(font,ed.gi))ed.gi--;
		}
		ei (c == KEY_END) {
			if (ed.gi < max_glyph-1)ed.gi++;
			while (ed.gi < max_glyph-1 && !get_glyph(font,ed.gi))ed.gi++;
		}
		ei (c == 'd') edit_glyph_pixel(get_glyph(font,ed.gi),ed.y,ed.x,1);
		ei (c == 'e') edit_glyph_pixel(get_glyph(font,ed.gi),ed.y,ed.x,0);
		else ed.redraw = 0;
	}
	printf("\e[7;1H");
	printf(CLRSCR TOPLEF BBLU "%s - unloading\n\r" CLRCOL,bdfeditlogo);
	free_bdffont(font);
        unload_unidat();
	unload_eawdat();
	oio_finish();
	return 0;
}

void quik_help(){
	printf(BRED
		"[^X] exit [^Q] svxt [^G] goto [^H] hist [F1] help \r\n"
		"[^C] copy [^V] past [^W] swap [^A] add  [^K] del" CLRCOL);
}

void show_help() {
	printf("\033[2J\033[HBDF EDIT - Help Screen\r\n");
	printf("[^Q] exit and save\n\r");
	printf("[^X] exit and do not save\n\r");
	printf("[^G] go to character code\n\r");
	printf("[^H]/[H] go to character from history\n\r");
	printf("[^C] copy character\n\r");
	printf("[^V] paste character\n\r");
	printf("[^W] swap character\n\r");
	printf("[^A] add character\n\r");
	printf("[^K] del character\n\r");
	printf("[pgup] prev character\n\r");
	printf("[pgdn] next character\n\r");
	printf("[home] prev drawn character\n\r");
	printf("[end] next drawn character\n\r");
	printf("←↑→↓ move around the editing field\n\r");
	printf("[j|i|l|k] shift around the editing field\n\r");
	printf("[d] draw at cursor\n\r");
	printf("[e] erase at cursor\n\r");
	printf("[ ] toggle at cursor\n\r");
	printf("[F1] get help\n\r");
	printf("[enter] leave help\n\r");
	while(1) {
		int c = oio_getkey();
		if(c == CTRL('Q'))break;
		if(c == CTRL('X'))break;
		if(c == '\r')break;
	}
}

