/*###########################################################################*\
 # UTF-8 parsing functions implementing "lenient UTF-8", i.e. any bytes      #
 # that do not parse as part of a UTF-8 character are treated as             #
 # single-byte characters, as if they were ascii. They can be identified if  #
 # needed using the utf8_classify function.                                  #
\*###########################################################################*/

/****************************************************************\
 * Utility functions for dealing with single Unicode codepoints *
\****************************************************************/
/* Get a string of the utf-8 representation of a unicode code point.
   application should copy this string before calling the function again. */
char *u8chtostr(unsigned ch);
/* return the length of ch if it were in utf-8 */
int u8chlen(unsigned ch);
/* Write unicode codepoint as utf-8 to a file. */
int u8fputch(FILE *f, unsigned ch);
int u8putch(unsigned ch); /* to stdout */
/* Get the onscreen width of a codepoint, approximately. */
int u8apxwid(unsigned ch);
/* Get the first unicode codepoint in string s */
unsigned u8strtoch(const char *s);
/* Get the ith unicode codepoint in string s of length n bytes */
unsigned u8ith(const char *s, int i, int n);
unsigned u8sith(const char *s, int i); // for zero-terminated string

/***************************************************************************\
 * u8scanch family: Get a unicode codepoint from utf-8 source. EOF at end. *
\***************************************************************************/
int u8_scanch(int (*getc)(void *f),void (*ungetc)(char,void*),void*f);
// abstract function
int u8fscanch(FILE *f); // from a FILE*
int u8scanch(); // from stdin
int u8sscanch(const char *s,const char **end);// C zero-terminated string
int u8memscanch(const char *s,int i,int len,int *end);
// memory block with byte index i

/**********************************************************\
 * Navigation functions for working within a utf-8 string *
\**********************************************************/
/* go from byte index i to the start of the previous utf-8 character
     (i can be at end of string). at beginning of string returns -1 */
int u8prevch(const char *s, int i);
/* go from byte index i to the start of the current utf-8 character
   (the one which includes byte i. therefore i cannot be at end of string) */
int u8resync(const char *s, int i);
/* go from byte index i to the start of the next utf-8 character,
   at end of string returns -1. */
int u8nextch(const char *s, int i, int n);
int u8snextch(const char *s, int i); // for null-terminated string
/* count the utf-8 codepoints in string s */
int u8chrcnt(const char *s, int n);
int u8schrcnt(const char *s); // for null-terminated string
/* get the byte index of the ith codepoint in the utf-8 string s */
int u8ithchr(const char *s, int i, int n);
int u8sithchr(const char *s, int i); // for null-terminated string
/* return the utf-8 decoded character at byte index i
   in string s. returns UTF8_EOSTR at end of string */
unsigned u8rdnxchr(const char *s, int i, int n);
unsigned u8srdnxchr(const char *s, int i); //for null-terminated string
/* return a bitfield classifying the utf-8 character located at byte i
   in string s of length n bytes.  */
unsigned u8classify(const char *s, int i, int n);
unsigned u8sclassify(const char *s, int i); // for null-terminated string

/* bitfield members from utf8_classify */
#define UTF8CLASS_ASCII 01
#define UTF8CLASS_2BYTE 02
#define UTF8CLASS_3BYTE 04
#define UTF8CLASS_4BYTE 010
#define UTF8CLASS_CONTI 020
#define UTF8CLASS_START 040
#define UTF8CLASS_VALID 0100
#define UTF8CLASS_INVAL 0200
#define UTF8CLASS_EOSTR 0400
/* value returned by readnext when at end */
#define UTF8_EOSTR (~0U)
